//	FBBase.h
//
//	Tetsuo Yamada
//
//	FakeBar's base.
//	defining Message constants, globals constants, 
//	settings structure, global functions.

#ifndef		FB_BASE_h_
#define		FB_BASE_h_

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<app/Message.h>
#include	<app/Roster.h>
#include	<interface/Alert.h>
#include	<kernel/fs_info.h>
#include	<storage/Directory.h>
#include	<storage/File.h>
#include	<storage/FindDirectory.h>
#include	<support/Beep.h>


//-----------------------------------------------------------
//	preference files hierarchy
//-----------------------------------------------------------
//	~/config/
//	settings/
//	FakeBar/
//	layers/, 	settings, 	targets,	shelf
//	<<layer files>>
//-----------------------------------------------------------


//===========================================================
//	Global Constants for Application
//===========================================================
extern const char *fb_app_name ;
extern const char *fb_signature ;
extern const char *fb_settings_name ;
extern const char *fb_layers_name ;
extern const char *fb_targets_name ;


//===========================================================
//	Constants for Messaging
//===========================================================
enum {

//-----------------------------------------------------------
//	messages for LayerInfo
//-----------------------------------------------------------

//	initialize LayerInfo.
//	contains no data.
	FB_INITIALIZE = 'fb\0\0', 
	
//	add items to the handler (LayerInfo). 
//	contains pointer named "items" as a ref items, if needed.
//	contains int32 named "index" as an item's index, if needed.
//	contains bool named "finish" as a notification of the end of loading.
//	contains cstr named "name" as a new layer name, if needed.
	FB_ITEMS_FOUND, 
	
//	check all items. handler is LayerInfo.
//	contains int32 named "index" as an index of item to check out.
	FB_CHECK_ITEMS, 

//-----------------------------------------------------------
//	messages for LayerManager
//-----------------------------------------------------------

//	make its layers stop/start working.
//	contains bool named "enable" as a flag.
	FB_ENABLE_LAYER, 
	
//	update settings.
//	contains archive of new settings, otherwise no data (then read from file).
	FB_UPDATE_SETTINGS, 
	
//	load layers from files...
//	contains no data.
	FB_LOAD_LAYERS, 

//-----------------------------------------------------------
//	messages for LayerBar
//-----------------------------------------------------------

//	end tracking.
//	contains no data.
/**** NOT USED ***********/
	FB_END_TRACKING, 

//-----------------------------------------------------------
//	messages for BarWindow
//-----------------------------------------------------------

//	switch layer. handler is Window()
//	contains cstr named "layer_name" as a name of the layer or
//	contains bool named "direction" as a switching direction
	FB_SWITCH_LAYER, 

//	switch workspace. handler is Window()
//	contains int32 named "workspace" as a index of the workspace
	FB_SWITCH_WORKSPACE, 
	
//	show/hide shelf.
//	contains bool named "show" as a flag to inform whether show or hide.
	FB_SWITCH_SHELF, 

//-----------------------------------------------------------
//	messages for AppMenu
//-----------------------------------------------------------

//	update icon.
//	contains no data.
	FB_UPDATE_ICON, 
	
//	modify items.
//	contains no data.
	FB_MODIFY_ITEMS, 
	
//-----------------------------------------------------------
//	messages for FBShelf
//-----------------------------------------------------------

//	clear all replicants.
//	contains no data.
	FB_CLEAR_REPLICANTS, 

//-----------------------------------------------------------
//	messages for IOSlave
//-----------------------------------------------------------
	
//	save layer. handler is io_slave
//	contains cstr named "name" as a layer file's name.
//	contains cstr named "paths" as items' path.
//	contains bool named "icon" as a flag to decide if create icon.
	FB_SAVE_LAYER, 
	
//	load layer. handler is io_slave
//	contains pointer named "layer" as a layer to load.
//	pointer named "looper" as a looper for the layer.
//	int32 named "blocksize" as a block size.
//	int32 named "start" as a first item index to load.
	FB_LOAD_LAYER, 
	
//	stop loading layer. handler is io_slave
//	contains pointer named "layer" as a layer to stop loading.
	FB_STOP_LOADING, 
	
//	delete layer.
//	contains cstr named "name" as a layer name.
	FB_DELETE_LAYER, 

//-----------------------------------------------------------
//	messages for PrefsWindow
//-----------------------------------------------------------

//	<for WindowSet>
	FB_WINDOW_MSG, 
	
//	<for WorkspaceSet>
	FB_WORKSPACE_MSG, 
	
//	<for LayerSet>
	FB_LAYER_NEW_LAYER, 
	FB_LAYER_SELECT_LAYER, 
	FB_LAYER_REMOVE_LAYER, 
	FB_LAYER_RENAME_LAYER, 
	FB_LAYER_REMOVE_ITEM, 
	
//	<for TargetSet>
	FB_TARGET_REMOVE_ITEM, 
	
//	<for MiscSet>
	FB_MISC_MSG, 
	
//	<for PrefsWindow>
//	revert.
//	no data.
	FB_REVERT_PREFS, 
	
//	set preferences.
//	no data.
	FB_SET_PREFS, 
	
//	ListView item's selection changed.
//	no data.
	FB_SELECTION_CHANGED, 
	
//	show open panel.
//	no data.
	FB_SHOW_OPEN_PANEL, 
	
//	add a new layer or remove a layer.
//	contains cstr named "name" as a layer name.
//	contains bool named "create" as a flag to make decision.
	FB_HANDLE_LAYER, 

//-----------------------------------------------------------
//	messages for be_app
//-----------------------------------------------------------
	
//	reload settings, layers, targets from preferences.
//	contains no data.
	FB_LOAD_PREFS, 

//	open file. handler is be_app
//	entry_ref named "entries" as a target file.
//	mime named "text/plain" as a dopped text.
//	entry_ref named "file" as a layer item (app).
	FB_OPEN, 

//	open with. handler is be_app
//	cstr named "applications" as a selected app's signature.
//	entry_ref named "refs" as a selected app's ref.
//	entry_ref named "file" as a target file.
	FB_OPEN_WITH, 

//	open parent. handler is be_app.
//	entry_ref named "file" as a layer item (app).
	FB_OPEN_PARENT, 
	
//	show dialog which shows its state to the user. handler is be_app.
//	contains pointer named "layer" as an invoked layer.
//	contains cstr named "path" as an item's path.
	FB_SHOW_DIALOG, 
	
//	show prefereces window. handler is be_app.
//	contains no data.
	FB_SHOW_PREFS, 
	
//	notification of closing PrefsWindow. handler is be_app.
//	contains no data.
	FB_PREFS_CLOSED, 
	
	FB_END_ENUM
} ;



//===========================================================
//	struct settings
//===========================================================
typedef enum bar_location {
	FB_LEFT_TOP = 0, 
	FB_RIGHT_TOP, 
	FB_RIGHT_BOTTOM, 
	FB_LEFT_BOTTOM
} ;

typedef enum bar_direction {
	FB_TOP_TO_BOTTOM = 0, 
	FB_BOTTOM_TO_TOP, 
	FB_LEFT_TO_RIGHT, 
	FB_RIGHT_TO_LEFT
} ;

typedef enum mouse_operation {
	FB_NO_USE = 0, 
	FB_PRIMARY_DOWN, 
	FB_SECONDARY_DOWN, 
	FB_DOUBLE_CLICK
} ;

enum {
	FB_READ_ALERT = 0, 		// error on reading
	FB_WRITE_ALERT, 		// error on writing
	FB_LAUNCH_ALERT, 		// error on launching
	FB_READ_LAYER_ALERT, 	// found an illegal layer file
	FB_INSTALL_ALERT 		// error on installing
} ;

// action flags
#define		FB_MINIMIZE			0x1		// minimize when (window is) not active
#define		FB_AUTOFOCUS		0x2		// FFM
#define		FB_DRAW_LABEL		0x4
#define		FB_DROP_EXTENTION	0x8		// NOT USED
#define		FB_ORIGINAL_ICON	0x10	// used when create a new layer file
#define		FB_WORKSPACE_CMD	0x20	// in the Application Menu...
#define		FB_ENABLE_SHELF		0x40	// if shows shelf
#define		FB_SHELF_BOTTOM		0x80	// install the shelf on bottom
#define		FB_DRAW_BORDER		0x100	// draw border around each item


struct settings {
/* variables */
	bar_location			location ;
	bar_direction			direction ;
	mouse_operation			forInvoke ;
	mouse_operation			forExpand ;
	uint32					flags ;
	icon_size				which ;
	uint32					workspaces ;
	int32					shelfSize ;
	char					*layers[32] ;
	
/* functions */
							settings () ;
							~settings () ;
	
	status_t				archive_to (BMessage *msg) const ;
	status_t				set_to (BMessage *msg) ;
	void					set_to (const settings& src) ;
	
	void					get_item_size (float *width, float *height) const ; // inlined
	void					get_item_margins (float *hor, float *ver) const ; // inlined
	
	void					set_flags (uint32 bits, bool on) ; // inlined
	
	bool					is_enabled_in (int32 workspace) const ; // inlined
	
	void					set_layer (const char *name, int32 workspace) ; // inlined
	
	/*	this function returns 0, if compared settings are the same.
		otherwise, returns the level of the differency.
		level 1: only performance
		level 2: invalidate views and need to reshape window
		level 3: switch layer
		level 4: hide or show window
		level 5: rebuild views...
		 */
	int32					compare (const settings *alt, 
									bool compareAllWorkspaces = false) const ;
} ;

//===========================================================
//	definitions
//===========================================================

#define		IS_COLUMN		(fb_settings->direction<2)
#define		IS_POSITIVE		(fb_settings->direction%2==0)

//	constants for layout GUIs
#define		BIG_SPACE		10
#define		SMALL_SPACE		5
#define		PREFS_MIN_WIDTH		500
#define		PREFS_MIN_HEIGHT	320

#define		FB_UNSHOWN_POINT	BPoint (-2000, -2000)

//	these values are not used currently...
#define		OB_INTERVAL			130*1000
#define		SHOW_INTERVAL		220*1000
#define		SELECT_INTERVAL		200*1000

//===========================================================
//	for alert
//===========================================================

void						alert (int32 type, 
									const char *msg, 
									status_t err) ;

//===========================================================
//	for settings files
//===========================================================
//	fb_app_name				"FakeBar" -> get home/config/settings/FakeBar
//	fb_settings_name		"settings" -> get settings/FakeBar/settings
//	fb_layers_name			"layers" -> get settings/FakeBar/layers
//	fb_targets_name			"targets" are available. -> settings/FakeBar/targets
//	fb_shelf_name			"shelf" -> get settings/FakeBar/shelf
status_t					get_settings (const char *name, entry_ref *ref) ;

void						init_settings () ;

status_t					get_settings_node (BNode *node) ;



//-----------------------------------------------------------
//	window_settings inline implementation
//-----------------------------------------------------------
inline void settings::get_item_size (float *width, float *height) const {
	*width = *height = (which==B_LARGE_ICON) ? 33 : 15 ;
	if (flags & FB_DRAW_LABEL) {
		if (which==B_LARGE_ICON) {
			*height = 45 ;
			*width = 53 ;
		} else {
			*width = 96 ;
		}
	}
} ;

inline void settings:: get_item_margins (float *hor, float *ver) const {
	*hor = *ver = 2 ;
	if (which==B_MINI_ICON && (flags & FB_DRAW_LABEL)) *hor = 10 ;
} ;

inline void settings:: set_layer (const char *name, int32 workspace) {
	if (workspace<0 || workspace>=32) return ;
	delete[] layers[workspace] ;
	layers[workspace] = new char [strlen (name) + 1] ;
	strcpy (layers[workspace], name) ;
} ;

inline void settings:: set_flags (uint32 bits, bool on) {
	flags = (on) ? (flags | bits) : (flags & ~bits) ;
} ;

inline bool settings:: is_enabled_in (int32 workspace) const {
	return ((0x01 << workspace) & workspaces) ;
} ;


#endif

