//	DirectoryMenu.cpp
//
//	Tetsuo Yamada

#include	"DirectoryMenu.h"


//===========================================================
//	class DirectoryMenu
//===========================================================
DirectoryMenu:: DirectoryMenu (const BDirectory *dir, menu_layout layout) 
			: BMenu (B_EMPTY_STRING, layout) {
	status_t err = B_OK ;
	BEntry entry ;
	
	err = dir->InitCheck () ;
	if (err!=B_OK) goto Finish ;
	
	err = dir->GetEntry (&entry) ;
	if (err!=B_OK) goto Finish ;
	
	err = entry.InitCheck () ;
	if (err!=B_OK) goto Finish ;
	
	if (!entry.Exists()) {
		err = B_ENTRY_NOT_FOUND ;
		goto Finish ;
	}
	
	err = entry.GetRef (&fDirRef) ;
	if (err!=B_OK) goto Finish ;
	
	SetName (fDirRef.name) ;
	
Finish : 
	fDir.Unset () ;
	fIsRoot = true ;
	
	SetTriggersEnabled (true) ;
	SetRadioMode (false) ;
	SetLabelFromMarked (false) ;
	SetFont (be_plain_font) ;
	SetName (B_EMPTY_STRING) ;
} ;

DirectoryMenu:: ~DirectoryMenu () {
	// do nothing
} ;

void DirectoryMenu:: RemoveAllItems () {
	RemoveItems (0, CountItems(), true) ;
} ;

FileMenuItem* DirectoryMenu:: StartHierarchy (const entry_ref *ref, 
				BMessage *message) {
	status_t err = B_OK ;
	BEntry entry ;
	BDirectory dir ;
	DirectoryMenu *submenu = NULL ;
	
	err = entry.SetTo (ref, true) ;
	if (err!=B_OK) goto AddIt ;
	
	err = entry.InitCheck () ;
	if (err!=B_OK) goto AddIt ;
	
	if (!entry.Exists()) {
		err = B_ENTRY_NOT_FOUND ;
		goto AddIt ;
	}
	
	if (!entry.IsDirectory()) {
		err = B_NOT_A_DIRECTORY ;
		goto AddIt ;
	}
	
	err = dir.SetTo (&entry) ;
	if (err!=B_OK) goto AddIt ;
	
	submenu = new DirectoryMenu (&dir) ;
	
AddIt : 
	FileMenuItem *item = NULL ;
	BMessage *msg = NULL ;
	if (message) {
		msg = new BMessage (*message) ;
		if (msg->HasRef ("entries")) msg->ReplaceRef ("entries", ref) ;
		else msg->AddRef ("entries", ref) ;
	}
	if (submenu) {
		item = new FileMenuItem (ref, submenu, msg) ;
	} else {
		item = new FileMenuItem (ref, msg) ;
	}
	
	return item ;
} ;

FileMenuItem* DirectoryMenu:: StartHierarchy (const char *path, 
				BMessage *message) {
	status_t err = B_OK ;
	BEntry entry ;
	BDirectory dir ;
	DirectoryMenu *submenu = NULL ;
	entry_ref ref ;
	
	err = get_ref_for_path (path, &ref) ;
	if (err!=B_OK) goto AddIt ;
	
	err = entry.SetTo (&ref, true) ;
	if (err!=B_OK) goto AddIt ;
	
	err = entry.InitCheck () ;
	if (err!=B_OK) goto AddIt ;
	
	if (!entry.Exists()) {
		err = B_ENTRY_NOT_FOUND ;
		goto AddIt ;
	}
	
	if (!entry.IsDirectory()) {
		err = B_NOT_A_DIRECTORY ;
		goto AddIt ;
	}
	
	err = dir.SetTo (&entry) ;
	if (err!=B_OK) goto AddIt ;
	
	submenu = new DirectoryMenu (&dir) ;
	
AddIt : 
	FileMenuItem *item = NULL ;
	BMessage *msg = NULL ;
	if (message) {
		msg = new BMessage (*message) ;
		if (msg->HasRef ("entries")) msg->ReplaceRef ("entries", &ref) ;
		else msg->AddRef ("entries", &ref) ;
	}
	if (submenu) {
		item = new FileMenuItem (path, submenu, msg) ;
	} else {
		item = new FileMenuItem (path, msg) ;
	}
	
	return item ;
} ;

bool DirectoryMenu:: AddDynamicItem (add_state state) {
	status_t err = B_OK ;
	int32 count = 0 ;
	
	switch (state) {
		case B_INITIAL_ADD : 
			if (fIsRoot) {
				//	remove all items to update.
				RemoveAllItems() ;
				fDir.Unset () ;
			}
			count = CountItems() ;
			
			err = fDir.InitCheck () ;
			if (count==0 && err!=B_OK) {
				// need to start dynamically adding.
				fDir.SetTo (&fDirRef) ;
				return true ;
			} else if (err==B_OK) {
				// aborted on adding, and now add again.
				return true ;
			} else {
				// already added.
				return false ;
			}
		case B_PROCESSING : 
			return AddNextItem () ;
			
		case B_ABORT : 
			return false ;
	}
	return false ;
} ;

bool DirectoryMenu:: AddNextItem () {
	FileMenuItem *fitem = NULL ;
	BMenuItem *mitem = NULL ;
	DirectoryMenu *submenu = NULL ;
	entry_ref ref ;
	
	switch (fDir.GetNextRef (&ref)) {
		case B_OK : 
			mitem = Superitem() ;
			fitem = StartHierarchy (&ref, mitem->Message()) ;
			if (fitem->Submenu()) {
				submenu = cast_as (fitem->Submenu(), DirectoryMenu) ;
				submenu->fIsRoot = false ;
			}
			fitem->SetTarget (mitem->Target()) ;
			AddItem (fitem) ;
			return true ;
		
		case B_ENTRY_NOT_FOUND : 
			if (CountItems()==0) {
				// empty folder
				mitem = new BMenuItem ("Empty Folder", (BMessage*)NULL) ;
				mitem->SetEnabled (false) ;
				AddItem (mitem) ;
			}
			fDir.Unset () ;
			break ;
			
		default : 
			mitem = new BMenuItem ("Internal Error Occurred", (BMessage*)NULL) ;
			mitem->SetEnabled (false) ;
			AddItem (mitem) ;
			fDir.Unset () ;
			break ;
	}
	return false ;
} ;

status_t DirectoryMenu:: SetDirectory (const BDirectory *dir) {
	status_t err = B_OK ;
	BEntry entry ;
	
	if (fDir.InitCheck()!=B_OK) {
		if (Window()) return B_INTERRUPTED ;
	} else {
		int32 count = CountItems() ;
		RemoveItems (0, count, true) ;
	}
	
	err = dir->GetEntry (&entry) ;
	if (err!=B_OK) return err ;
	
	err = entry.InitCheck() ;
	if (err!=B_OK) return err ;
	
	if (!entry.Exists()) return B_ENTRY_NOT_FOUND ;
	
	err = entry.GetRef (&fDirRef) ;
	if (err!=B_OK) return err ;
	
	if (fDir.InitCheck()==B_OK) {
		fDir = *dir ;
		err = fDir.InitCheck () ;
		if (err!=B_OK) return err ;
	}
	
	return B_OK ;
} ;

status_t DirectoryMenu:: GetDirectory (BDirectory *dir) const {
	status_t err = B_OK ;
	
	err = dir->SetTo (&fDirRef) ;
	if (err!=B_OK) return err ;
	
	return B_OK ;
} ;