//	MatrixListView.cpp
//
//	Tetsuo Yamada

#include	"MatrixListView.h"


//===================================================================
//	class ListColumn
//===================================================================
ListColumn:: ListColumn (uint32 c, const char *name, float width) 
		: BArchivable () {
	command = c ;
	fFrame = BRect (0, 0, width, 10000) ;
	fName = new char [strlen (name) + 1] ;
	strcpy (fName, name) ;
	fAlign = B_ALIGN_LEFT ;
} ;

ListColumn:: ~ListColumn () {
	delete[] fName ;
} ;

float ListColumn:: PreferredWidth () const {
	if (!fOwner) return fFrame.Width () ;
	int32 count = fOwner->CountItems () ;
	float max = 0 ;
	MatrixListItem *item = NULL ;
	for (int32 i=0 ; i<count ; i++) {
		item = (MatrixListItem*)(fOwner->ItemAt (i)) ;
		max = max_c (max, item->MinimumColumnWidth (this)) ;
	}
	return max ;
} ;

void ListColumn:: ResizeToPreferred () {
	if (!fOwner) return ;
	SetWidth (PreferredWidth ()) ;
} ;

void ListColumn:: SetWidth (float width) {
	float gap = width - fFrame.Width() ;
	fFrame.right = fFrame.left + width ;
	if (fOwner && gap!=0) {
		int32 index = fOwner->ColumnIndexOf (this) ;
		int32 count = fOwner->CountColumns () ;
		for (int32 i=index + 1 ; i<count ; i++) {
			fOwner->ColumnAt (i)->fFrame.OffsetBy (gap, 0) ;
		}
		
		BRect bounds = fOwner->Bounds() ;
		bounds.left = fFrame.left ;
		fOwner->Invalidate (bounds) ;
	}
} ;

BRect ListColumn:: Frame () const {
	return fFrame ;
} ;

float ListColumn:: Width () const {
	return fFrame.Width() ;
} ;

const char* ListColumn:: Name () const {
	return fName ;
} ;

void ListColumn:: SetName (const char *name) {
	delete[] fName ;
	name = new char [strlen (name) + 1] ;
	strcpy (fName, name) ;
} ;

alignment ListColumn:: Align () const {
	return fAlign ;
} ;

void ListColumn:: SetAlign (alignment align) {
	bool changed = (fAlign!=align) ;
	fAlign = align ;
	
	if (changed && fOwner) {
		BRect rect = fOwner->Bounds() ;
		rect.left = fFrame.left ;
		rect.right = fFrame.right ;
		fOwner->Invalidate (rect) ;
	}
} ;

//===================================================================
//	class MatrixListView
//===================================================================
MatrixListView:: MatrixListView (BRect frame, 
				const char *name, 
				uint32 informFlags, 
				list_view_type type, 
				uint32 mode, 
				uint32 flags)
		: BListView (frame, name, type, mode, flags) {
	fGap = 4 ;
	fUnderlined = -1 ;
	fFlags = informFlags ;
} ;

MatrixListView:: ~MatrixListView () {
	int32 count = fColumns.CountItems () ;
	for (int32 i=count - 1 ; i>=0 ; i--) {
		delete fColumns.RemoveItem (i) ;
	}
} ;

void MatrixListView:: AttachedToWindow () {
	BListView::AttachedToWindow () ;
	SetScrollBars () ;
} ;

void MatrixListView:: FrameResized (float width, float height) {
	BListView::FrameResized (width, height) ;
	SetScrollBars () ;
} ;

void MatrixListView:: GetPreferredSize (float *width, float *height) {
	BListView::GetPreferredSize (width, height) ;
	*width = 0 ;
	int32 count = CountColumns () ;
	ListColumn *column = NULL ;
	for (int32 i=0 ; i<count ; i++) {
		column = (ListColumn*)(fColumns.ItemAt (i)) ;
		*width = 2*fGap + column->PreferredWidth () ;
	}
} ;

void MatrixListView:: MouseMoved (BPoint p, uint32 t, const BMessage *msg) {
	if (!msg || msg->what!='DATA') {
		BListView::MouseMoved (p, t, msg) ;
		return ;
	}
	
	int32 index = IndexOf (p) ;
	if (index==-1) index = CountItems () ;
	switch (t) {
		case B_EXITED_VIEW :
			if (fUnderlined>-1) {
				InvalidateItem (fUnderlined) ;
				fUnderlined = -1 ;
			}
			break ;
		case B_INSIDE_VIEW :
		case B_ENTERED_VIEW :
			UpdateUnderline (index) ;
			break ;
	}
	
	// scroll view vertically as the user wants!
	BRect bounds = Bounds () ;
	uint32 buttons = B_PRIMARY_MOUSE_BUTTON ;
	float b = ItemFrame (CountItems() - 1).bottom ;
	
	if (p.y<bounds.top + 15 && bounds.top>0) {
		while (buttons && p.y<bounds.top + 15 && bounds.top>0) {
			if (p.y<bounds.top) ScrollTo (bounds.left, max_c (bounds.top - 8, 0)) ;
			else ScrollTo (bounds.left, max_c (bounds.top - 4, 0)) ;
			UpdateUnderline (index) ;
			
			snooze (20*1000) ;
			GetMouse (&p, &buttons) ;
			bounds = Bounds() ;
			index = IndexOf (p) ;
			if (index==-1) index = CountItems () ;
		}
	} else if (p.y>bounds.top + 15 && bounds.bottom<b) {
		while (buttons && p.y>bounds.bottom - 15 && bounds.bottom<b) {
			if (p.y>bounds.bottom) ScrollTo (bounds.left, bounds.top + 8) ;
			else ScrollTo (bounds.left, bounds.top + 4) ;
			UpdateUnderline (index) ;
			
			snooze (20*1000) ;
			GetMouse (&p, &buttons) ;
			bounds = Bounds () ;
			index = IndexOf (p) ;
			if (index==-1) index = CountItems () ;
		}
	}
} ;

void MatrixListView:: MessageReceived (BMessage *msg) {
	if (msg->WasDropped()) {
		void *pointer = NULL ;
		if (msg->FindPointer ("p:mlv_source", &pointer)==B_OK && pointer==this) {
			HandleItemDropping (msg) ;
		} else if (fFlags & P_MLV_MSG_DROPPED) {
			BPoint p = ConvertFromScreen (msg->DropPoint()) ;
			BMessage message (P_MLV_MSG_DROPPED) ;
			message.AddMessage ("dropped_msg", msg) ;
			message.AddPoint ("dropped_point", p) ;
			message.AddInt32 ("dropped_index", IndexOf (p)) ;
			Invoke (&message) ;
		}
		
		if (fUnderlined!=-1) {
			InvalidateItem (fUnderlined) ;
			fUnderlined = -1 ;
		}
	} else {
		BListView::MessageReceived (msg) ;
	}
} ;

bool MatrixListView:: InitiateDrag (BPoint p, int32 index, bool selected) {
	if (!selected) return false ;
	
	BMessage msg ('DATA') ;
	status_t err = B_OK ;
	int32 i = -1, j = 0 ;
	MatrixListItem *anItem = NULL ;
	BRect rect = ItemFrame (index) ;
	
	while ((i = CurrentSelection (j++))>=0) {
		anItem = (MatrixListItem*)(ItemAt (i)) ;
		
		err = anItem->MessageForDrag (&msg) ;
		if (err!=B_OK) continue ;
		
		err = msg.AddInt32 ("p:mlv_item_index", i) ;
		if (err!=B_OK) continue ;
		
		rect = (rect | ItemFrame (i)) ;
	}
	
	err = msg.AddPointer ("p:mlv_source", this) ;
	if (err!=B_OK) return false ;
	
	DragMessage (&msg, rect) ;
	return true ;
} ;

bool MatrixListView:: AddItem (BListItem *item, int32 index) {
	MatrixListItem *anItem = cast_as (item, MatrixListItem) ;
	if (!anItem) return false ;
	
	bool r = BListView::AddItem (item, index) ;
	if (r && (fFlags & P_MLV_ITEM_ADDED)) {
		BMessage msg (P_MLV_ITEM_ADDED) ;
		msg.AddInt32 ("index", index) ;
		msg.AddPointer ("item", &item) ;
		Invoke (&msg) ;
	}
	return r ;
} ;

bool MatrixListView:: AddItem (BListItem *item) {
	MatrixListItem *anItem = cast_as (item, MatrixListItem) ;
	if (!anItem) return false ;
	
	bool r = BListView::AddItem (item) ;
	if (r && (fFlags & P_MLV_ITEM_ADDED)) {
		BMessage msg (P_MLV_ITEM_ADDED) ;
		msg.AddInt32 ("index", CountItems() - 1) ;
		msg.AddPointer ("item", &item) ;
		Invoke (&msg) ;
	}
	return r ;
} ;

BListItem* MatrixListView:: RemoveItem (int32 index) {
	BListItem *item = BListView::RemoveItem (index) ;
	if (item && (fFlags & P_MLV_ITEM_REMOVED)) {
		BMessage msg (P_MLV_ITEM_REMOVED) ;
		msg.AddInt32 ("index", index) ;
		msg.AddPointer ("item", &item) ;
		Invoke (&msg) ;
	}
	return item ;
} ;

bool MatrixListView:: RemoveItem (BListItem *item) {
	int32 index = IndexOf (item) ;
	bool r = BListView::RemoveItem (item) ;
	if (r && (fFlags & P_MLV_ITEM_REMOVED)) {
		BMessage msg (P_MLV_ITEM_REMOVED) ;
		msg.AddInt32 ("index", index) ;
		msg.AddPointer ("item", &item) ;
		Invoke (&msg) ;
	}
	return r ;
} ;

bool MatrixListView:: AddColumn (ListColumn *column, int32 index) {
	bool result = column ;
	ListColumn *prev = NULL ;
	if (index==-1) {
		prev = (ListColumn*)(fColumns.LastItem()) ;
		result = (result && fColumns.AddItem (column)) ;
	} else {
		if (index!=0) prev = (ListColumn*)(fColumns.ItemAt (index - 1)) ;
		result = (result && fColumns.AddItem (column, index)) ;
	}
	
	if (result) {
		column->fOwner = this ;
		
		float width = column->fFrame.Width() ;
		column->fFrame.left = (prev) ? prev->fFrame.right + 1 : 0 ;
		column->fFrame.right = column->fFrame.left - 1 ;
		column->SetWidth (width) ;
		
		if (fFlags & P_MLV_COLUMN_ADDED) {
			BMessage msg (P_MLV_COLUMN_ADDED) ;
			msg.AddInt32 ("index", (index==-1) ? CountItems() - 1 : index) ;
			msg.AddPointer ("part", &column) ;
			Invoke (&msg) ;
		}
		
		SetScrollBars () ;
	}
	return result ;
} ;

bool MatrixListView:: RemoveColumn (ListColumn *column) {
	int32 index = ColumnIndexOf (column) ;
	return (index==-1) ? false : RemoveColumn (index) ;
} ;

ListColumn* MatrixListView:: RemoveColumn (int32 index) {
	ListColumn *column = (ListColumn*)(fColumns.RemoveItem (index)) ;
	if (column) {
		ListColumn *next = (ListColumn*)(fColumns.RemoveItem (index)) ;
		if (next) {
			float width = next->fFrame.Width() ;
			next->fFrame.left = column->fFrame.left ;
			next->SetWidth (width) ;
		}
		
		column->fOwner = NULL ;
		column->fFrame.Set (0, 0, column->fFrame.Width(), 10000) ;
		
		if (fFlags & P_MLV_COLUMN_REMOVED) {
			BMessage msg (P_MLV_COLUMN_REMOVED) ;
			msg.AddInt32 ("index", index) ;
			msg.AddPointer ("part", &column) ;
			Invoke (&msg) ;
		}
		SetScrollBars () ;
	}
	return column ;
} ;

int32 MatrixListView:: CountColumns () const {
	return fColumns.CountItems () ;
} ;

ListColumn* MatrixListView:: ColumnAt (int32 index) const {
	return (ListColumn*)(fColumns.ItemAt (index)) ;
} ;

ListColumn* MatrixListView:: ColumnAt (BPoint p) const {
	if (p.x<0) return NULL ;
	int32 count = fColumns.CountItems () ;
	ListColumn *column = NULL ;
	for (int32 i=0 ; i<count ; i++) {
		column = (ListColumn*)(fColumns.ItemAt (i)) ;
		if (column->fFrame.Contains (p)) return column ;
	}
	return NULL ;
} ;

int32 MatrixListView:: ColumnIndexOf (ListColumn *column) const {
	return fColumns.IndexOf (column) ;
} ;

bool MatrixListView:: MoveColumn (int32 from, int32 to) {
	int32 flags = fFlags ;
	fFlags = 0 ;
	bool result = false ;
	ListColumn *column = RemoveColumn (from) ;
	
	if (!column) {
		result = false ;
		goto End ;
	}
	
	result = AddColumn (column, to) ;
	if (result && (flags & P_MLV_COLUMN_MOVED)) {
		BMessage msg (P_MLV_COLUMN_MOVED) ;
		msg.AddInt32 ("from", from) ;
		msg.AddInt32 ("to", to) ;
		msg.AddPointer ("item", &column) ;
		Invoke (&msg) ;
	}
	
End :
	fFlags = flags ;
	return result ;
} ;

void MatrixListView:: SetInformFlags (uint32 flags) {
	fFlags = flags ;
} ;

uint32 MatrixListView:: InformFlags () const {
	return fFlags ;
} ;

void MatrixListView:: SetGap (float gap) {
	fGap = gap ;
} ;

float MatrixListView:: Gap () const {
	return fGap ;
} ;

void MatrixListView:: UpdateUnderline (int32 index) {
	if (index - 1==fUnderlined) {
		return ;
	} else if (index>0) {
		if (fUnderlined>=0) InvalidateItem (fUnderlined) ;
		rgb_color high = HighColor () ;
		BRect rect = ItemFrame (index - 1) ;
		rect.top = rect.bottom - 1 ;
		SetHighColor (tint_color (ui_color2 (P_STANDARD_GRAY), B_DARKEN_4_TINT)) ;
		FillRect (rect) ;
		SetHighColor (high) ;
		fUnderlined = index - 1 ;
	} else {
		if (fUnderlined>=0) InvalidateItem (fUnderlined) ;
		fUnderlined = -1 ;
	}
} ;

void MatrixListView:: HandleItemDropping (BMessage *msg) {
	BPoint p = ConvertFromScreen (msg->DropPoint()) ;
	BListItem *dropped = ItemAt (IndexOf (p)) ;
	int32 index ; // index of an item
	int32 count ;
	type_code type ;
	int32 flags = fFlags ;
	fFlags = 0 ;
	
	msg->GetInfo ("p:mlv_item_index", &type, &count) ;
	BListItem **items = new BListItem *[count] ;
	
	for (int32 i=0 ; i<count ; i++) {
		if (msg->FindInt32 ("p:mlv_item_index", i, &index)==B_OK)
			items[i] = ItemAt (index) ;
	}
	
	for (int32 i=0 ; i<count ; i++) {
		index = IndexOf (items[i]) ;
		RemoveItem (index) ;
		
		if (dropped && items[i]!=dropped) AddItem (items[i], IndexOf (dropped)) ;
		else if (dropped) AddItem (items[i], index) ;
		else AddItem (items[i]) ;
		
		if (flags & P_MLV_ITEM_MOVED) {
			BMessage message (P_MLV_ITEM_MOVED) ;
			message.AddInt32 ("from", index) ;
			message.AddInt32 ("to", IndexOf (items[i])) ;
			message.AddPointer ("item", &(items[i])) ;
			Invoke (&message) ;
		}
	}
	delete[] items ;
	fFlags = flags ;
} ;

void MatrixListView:: SetScrollBars () {
	int32 ycount = CountItems () ;
	int32 xcount = fColumns.CountItems() ;
	BScrollBar *bar = NULL ;
	BRect bounds = Bounds () ;
	BRect page ;
	if (ycount>0) {
		page = ItemFrame (ycount - 1) ;
		page.left = 0 ;
		page.top = 0 ;
		if (xcount>0) page.right = ColumnAt (xcount - 1)->Frame().right ;
	} else {
		page = bounds ;
	}
	
	if ((bar = ScrollBar (B_HORIZONTAL))) {
		if (bounds.Width()<page.right) {
			bar->SetProportion (bounds.Width()/page.right) ;
			bar->SetRange (0, page.right - bounds.Width()) ;
		} else {
			bar->SetProportion (1) ;
			bar->SetRange (0, 0) ;
		}
		bar->SetSteps (5, bounds.Width()) ;
	}
	
	if ((bar = ScrollBar (B_VERTICAL))) {
		if (bounds.Height()<page.bottom) {
			bar->SetProportion (bounds.Height()/page.bottom) ;
			bar->SetRange (0, page.bottom - bounds.Height()) ;
		} else {
			bar->SetProportion (1) ;
			bar->SetRange (0, 0) ;
		}
		bar->SetSteps (5, bounds.Height()) ;
	}
} ;

void MatrixListView:: ScrollTo (BPoint p) {
	SetScrollBars () ;
	BListView::ScrollTo (p) ;
} ;