//	LayerInfo.cpp
//
//	Tetsuo Yamada

#include	"LayerInfo.h"
#include	"IOSlave.h"
#include	"bitmap.h"


extern IOSlave* fb_slave ;
extern const char *fb_layers_name ;
extern const char *fb_targets_name ;
extern const char *fb_app_name ;


//==============================================================
//	globals for layer
//==============================================================

bool is_layer_file (const entry_ref *ref) {
	status_t err = B_OK ;
	entry_ref dir, pref ;
	BEntry entry, parent ;
	char type [B_MIME_TYPE_LENGTH + 1] ;
	
	err = get_type (ref, type, NULL, true) ;
	if (err<B_OK) return false ;
	
	if (strcmp (type, "text/plain")!=0) return false ;
	
	err = get_settings (fb_layers_name, &dir) ;
	if (err!=B_OK) return false ;
	
	err = entry.SetTo (ref) ;
	if (err!=B_OK) return false ;
	
	err = entry.InitCheck () ;
	if (err!=B_OK) return false ;
	
	err = entry.GetParent (&parent) ;
	if (err!=B_OK) return false ;
	
	err = parent.GetRef (&pref) ;
	if (err!=B_OK) return false ;
	
	return (pref==dir) ;
} ;

bool is_layer_file (const char *path) {
	
	status_t err = B_OK ;
	char type [B_MIME_TYPE_LENGTH + 1] ;
	char dir [B_PATH_NAME_LENGTH + 1] ;
	
	// check path
	err = find_directory (B_USER_SETTINGS_DIRECTORY, dev_for_path ("/boot"), 
			false, dir, B_PATH_NAME_LENGTH) ;
	if (err!=B_OK) return false ;
	
	strcat (dir, "/") ;
	strcat (dir, fb_app_name) ;
	strcat (dir, "/") ;
	strcat (dir, fb_layers_name) ;
	strcat (dir, "/") ;
	
	if (strncmp (dir, path, strlen (dir))!=0) return false ;
	
	// check file-type
	err = get_type (path, type, NULL, true) ;
	if (err!=B_OK) return false ;
	
	if (strcmp (type, "text/plain")!=0) return false ;
	
	return true ;
} ;

void write_layer_attr (BFile *file, bool icon) {
	// write file type
	file->WriteAttr ("BEOS:TYPE", B_MIME_TYPE, 0, "text/plain", 11) ;
	if (!icon) return ;
	
	// write icon infos
	file->WriteAttr ("BEOS:M:STD_ICON", B_COLOR_8_BIT_TYPE, 0, fb_mini_icon, 256) ;
	file->WriteAttr ("BEOS:L:STD_ICON", B_COLOR_8_BIT_TYPE, 
			0, fb_large_icon, 1024) ;
} ;

//==============================================================
//	class LayerInfo
//==============================================================
LayerInfo:: LayerInfo (const char *layerName, bool withFile) 
			: BHandler () {
	fData = 0 ;
	fName = new char [strlen (layerName) + 1] ;
	strcpy (fName, layerName) ;
	fLocker = &fData ;
	fCompleted = false ;
	fWithFile = withFile ;
} ;

LayerInfo:: ~LayerInfo () {
	delete[] fName ;
	int32 count = fItems.CountItems () ;
	RefRetriever *item = NULL ;
	for (int32 i=count - 1 ; i>=0 ; i--) {
		item = (RefRetriever*)(fItems.RemoveItem (i)) ;
		delete item ;
	}
} ;

const char* LayerInfo:: LayerName () const {
	return fName ;
} ;

void LayerInfo:: SetLayerName (const char *name) {
	delete[] fName ;
	fName = new char [strlen (name) + 1] ;
	strcpy (fName, name) ;
} ;

bool LayerInfo:: IsCompleted () const {
	return fCompleted ;
} ;

int32 LayerInfo:: CountItems () const {
	return fItems.CountItems () ;
} ;

void* LayerInfo:: Locker () const {
	return fLocker ;
} ;

int32 LayerInfo:: CompareItems (const LayerInfo* info) const {
	int32 tCount = fItems.CountItems () ;
	int32 iCount = info->fItems.CountItems () ;
	if (tCount!=iCount) return 1 ;
	
	RefRetriever *titem = NULL ;
	RefRetriever *iitem = NULL ;
	for (int32 i=min_c (tCount, iCount) - 1 ; i>=0 ; i--) {	
		titem = (RefRetriever*)(fItems.ItemAt (i)) ;
		iitem = (RefRetriever*)(info->fItems.ItemAt (i)) ;
		if (strcmp (titem->Path(), iitem->Path())!=0) return 1 ;
	}
	return 0 ;
} ;

BList* LayerInfo:: SetLock (bool lock, void* locker) {
	if (fLocker==&fData) return NULL ;
	
	if (fLocker==NULL && lock) { // lock layer
		fLocker = locker ;
		return &fItems ;
	} else if (fLocker==locker && !lock) { // unlock layer
		fLocker = NULL ;
		return NULL ;
	} else if (fLocker==locker && lock) { // layer locking object tries to lock again.
		return &fItems ;
	}
	
	return NULL ;
} ;

void LayerInfo:: SetEnabled (bool enabled) {
	
	if (enabled==(fLocker!=&fData)) return ;
	
	if (enabled) {
		int32 count = fItems.CountItems () ;
		
		if (!fCompleted && fWithFile) {
			// post message to load layer
			BMessage msg (FB_LOAD_LAYER) ;
			msg.AddPointer ("layer", this) ;
			msg.AddPointer ("looper", Looper()) ;
			msg.AddInt32 ("blocksize", 1) ;
			msg.AddInt32 ("start", count) ;
			
			fb_slave->PostMessage (&msg) ;
		}
		
		if (fLocker==&fData) fLocker = NULL ;
	} else {
		if (!fCompleted) {
			// stop loading
			BMessage msg (FB_STOP_LOADING) ;
			msg.AddPointer ("layer", this) ;
			
			fb_slave->PostMessage (&msg) ;
		}
		
		// now i am not available, so...
		fLocker = &fData ;
	}
} ;

void LayerInfo:: MessageReceived (BMessage *msg) {
	switch (msg->what) {
		case FB_INITIALIZE : 
			Initialize () ;
			break ;
		case FB_ITEMS_FOUND : 
			ItemsFound (msg) ;
			break ;
		case FB_CHECK_ITEMS : {
			if (!fCompleted) break ;
			int32 index = -1 ;
			
			msg->FindInt32 ("index", &index) ;
			if (index>=0 && index<CountItems()) 
				((RefRetriever*)(fItems.ItemAt (index)))->CheckNode (false) ;
			break ;
		}
		default : 
			BHandler::MessageReceived (msg) ;
			break ;
	}
} ;

void LayerInfo:: Initialize () {
	
	if (fLocker!=&fData) SetEnabled (false) ;
	int32 count = fItems.CountItems () ;
	RefRetriever *item = NULL ;
	for (int32 i=count - 1 ; i>=0 ; i--) {
		item = (RefRetriever*)(fItems.RemoveItem (i)) ;
		delete item ;
	}
	fCompleted = false ;
	fLocker = &fData ;
} ;

void LayerInfo:: ItemsFound (BMessage *msg) {
	
	status_t err = B_OK ;
	int32 index = 0 ;
	RefRetriever* item = NULL ;
	const char *name = NULL ;
	bool finish = false ;
	
	while (msg->FindPointer ("items", index++, (void **)&item)==B_OK) {
		fItems.AddItem (item) ;
	}
	
	err = msg->FindBool ("finish", &finish) ;
	if (err==B_OK && finish) fCompleted = finish ;
	
	err = msg->FindString ("name", &name) ;
	if (err==B_OK && strcmp (name, fName)!=0) {
		delete[] fName ;
		fName = new char [strlen (name) + 1] ;
		strcpy (fName, name) ;
	}
} ;

void LayerInfo:: CheckAllItems (bool deep) {
	int32 count = fItems.CountItems () ;
	RefRetriever* item = NULL ;
	for (int32 i=0 ; i<count ; i++) {
		item = (RefRetriever*)(fItems.ItemAt (i)) ;
		item->CheckNode (deep) ;
	}
} ;

void LayerInfo:: PrintToStream () const {
	char state [32] ;
	int32 count = fItems.CountItems() ;
	
	if (fLocker==NULL) {
		// layer is unlocked.
		strcpy (state, "is unlocked.") ;
	} else if (fLocker==&fData) {
		// layer is disabled.
		strcpy (state, "is disabled.") ;
	} else {
		// layer is locked.
		strcpy (state, "is locked.") ;
	}
	
	printf ("LayerInfo name=%s, count=%d, completed=%s, %s\n", 
			fName, count, fCompleted, state) ;
	for (int32 i=0 ; i<count ; i++) {
		RefRetriever* item = (RefRetriever*)(fItems.ItemAt (i)) ;
		printf ("item[%d]=\"%s\"\n", i, item->Path()) ;
	}
} ;

//==============================================================
//	class LayerManager
//==============================================================
void LayerManager:: Show () {
	BWindow::Show () ;
	
	Lock () ;
	if (!IsHidden()) {
		SetLayerEnabled (true) ;
	}
	Unlock () ;
} ;

void LayerManager:: Quit () {
	int32 count = CountHandlers () ;
	LayerInfo *info = NULL ;
	
	//	we must remove/delete BHandler(s) with our hands.
	for (int32 i=count - 1 ; i>=0 ; i--) {
		info = cast_as (HandlerAt (i), LayerInfo) ;
		if (!info) continue ;
		
		info->SetEnabled (false) ;
		RemoveHandler (info) ;
		delete info ;
	}
	
	BWindow::Quit () ;
} ;

void LayerManager:: MessageReceived (BMessage *msg) {
	bool flag = false ;
	status_t err = B_OK ;
	
	switch (msg->what) {
		case FB_ENABLE_LAYER : 
			if (msg->FindBool ("enable", &flag)==B_OK) SetLayerEnabled (flag) ;
			break ;
		case FB_UPDATE_SETTINGS :
			UpdateSettings (msg, true) ;
			break ;
		case FB_LOAD_LAYERS : 
			err = LoadLayers () ;
			if (err!=B_OK) alert (-1, "Failed to load layers...", err) ;
			break ;
	}
} ;

status_t LayerManager:: LoadLayers () {
	
	status_t err = B_OK ;
	BDirectory dir ;
	entry_ref ref ;
	char type [B_MIME_TYPE_LENGTH + 1] ;
	int32 count = CountHandlers () ;
	LayerInfo *info = NULL ;
	
	for (int32 i=count - 1 ; i>=0 ; i--) {
		info = cast_as (HandlerAt (i), LayerInfo) ;
		if (!info) continue ;
		
		info->SetEnabled (false) ;
		RemoveHandler (info) ;
		delete info ;
	}
	
	// first, get layers
	err = get_settings (fb_layers_name, &ref) ;
	if (err!=B_OK) return err ;
	
	err = dir.SetTo (&ref) ;
	if (err!=B_OK) return err ;
	
	err = dir.InitCheck () ;
	if (err!=B_OK) return err ;
	
	while (dir.GetNextRef (&ref)==B_OK) {
		err = get_type (&ref, type, NULL, true) ;
		if (err==B_OK && strcmp (type, "text/plain")==0) {
			AddLayer (ref.name) ;
		} else {
			alert (FB_READ_LAYER_ALERT, ref.name, B_OK) ;
		}
	}
	
	// then, get targets
	AddLayer (B_EMPTY_STRING) ;
	
	return B_OK ;
} ;

void LayerManager:: UpdateSettings (BMessage *msg, bool call) {
	status_t err = B_OK ;
	entry_ref ref ;
	BFile file ;
	settings s ;
	BMessage message ;
	
	// try to set settings from archive.
	if (msg && s.set_to (msg)==B_OK) goto Check ;
	
	// load settings file.
	err = get_settings (fb_settings_name, &ref) ;
	if (err==B_OK) err = file.SetTo (&ref, B_READ_ONLY) ;
	if (err==B_OK) err = file.InitCheck () ;
	if (err==B_OK) err = message.Unflatten (&file) ;
	if (err==B_OK) err = s.set_to (&message) ;
	
Check : 
	if (call) UpdateSettings (&s) ;
	else fSettings.set_to (s) ;
} ;

void LayerManager:: UpdateSettings (settings *s) {
	if (fSettings.compare (s, false)) {
		fSettings.set_to (*s) ;
	}
} ;

void LayerManager:: SetLayerEnabled (bool enabled) {
	int32 count = CountHandlers () ;
	LayerInfo *info = NULL ;
	for (int32 i=0 ; i<count ; i++) {
		info = cast_as (HandlerAt (i), LayerInfo) ;
		if (!info) continue ;
		
		info->SetEnabled (enabled) ;
	}
} ;

status_t LayerManager:: GetLayerList (BMessage *container) const {
	status_t err = B_OK ;
	LayerInfo *info = NULL ;
	int32 count = CountHandlers () ;
	
	for (int32 i=0 ; i<count ; i++) {
		info = cast_as (HandlerAt (i), LayerInfo) ;
		if (!info) continue ;
		
		err = container->AddString ("name", info->fName) ;
		if (err!=B_OK) return err ;
	}
		
	err = container->AddInt32 ("layers_which", fLayersWhich) ;
	if (err!=B_OK) return err ;
	
	err = container->AddInt32 ("targets_which", fTargetsWhich) ;
	if (err!=B_OK) return err ;
	
	return B_OK ;
} ;

status_t LayerManager:: GetLayerPointer (BMessage *container) const {
	status_t err = B_OK ;
	LayerInfo *info = NULL ;
	int32 count = CountHandlers () ;
	
	for (int32 i=0 ; i<count ; i++) {
		info = cast_as (HandlerAt (i), LayerInfo) ;
		if (!info) continue ;
		
		err = container->AddPointer (info->fName, info) ;
		if (err!=B_OK) return err ;
	}
	
	return B_OK ;
} ;

LayerInfo* LayerManager:: FindLayer (const char *name) const {
	
	int32 count = CountHandlers () ;
	LayerInfo *info = NULL ;
	
	for (int32 i=0 ; i<count ; i++) {
		info = cast_as (HandlerAt (i), LayerInfo) ;
		if (!info) continue ;
		
		if (strcmp (name, info->fName)==0) return info ;
	}
	
	return NULL ;
} ;

LayerInfo* LayerManager:: FindNextLayer (LayerInfo *start, int32 increment) const {
	
	if (increment==0) return start ;
	
	LayerInfo* info = NULL ;
	BHandler *handler = NULL ;
	int32 index = 0 ;
	int32 count = CountHandlers () ;
	
	if (start) index = IndexOf (start) + increment ;
	else if (increment>0) index = 0 ;
	else index = count - 1 ;
	
	if (index<0) return NULL ;
	
	while (true) {
		handler = HandlerAt (index) ;
		if (!handler) {
			if (increment>0) index = 0 ;
			else index = count - 1 ;
			handler = HandlerAt (index) ;
		}
		index += increment ;
		
		info = cast_as (handler, LayerInfo) ;
		if (!info) continue ;
		
		return info ;
	}
	return NULL ;
} ;

LayerManager:: LayerManager (BRect frame, const char *title, uint32 flags, 
					list_which layersWhich, list_which targetsWhich) 
			: BWindow (frame, title, B_TITLED_WINDOW, flags) {
	
	fLayersWhich = layersWhich ;
	fTargetsWhich = targetsWhich ;
	
	status_t err = B_OK ;
	UpdateSettings () ;
	err = LayerManager::LoadLayers () ;
	if (err!=B_OK) alert (-1, "Failed to load layers...", err) ;
} ;

LayerManager:: ~LayerManager () {
	// do nothing
} ;

LayerInfo* LayerManager:: AddLayer (const char *name, bool withFile) {
	
	LayerInfo *info = new LayerInfo (name, withFile) ;
	AddHandler (info) ;
	
	if (!IsHidden()) info->SetEnabled (true) ;
	return info ;
} ;

LayerInfo* LayerManager:: RemoveLayer (const char *name) {
	LayerInfo *info = FindLayer (name) ;
	if (!info) return NULL ;
	
	info->SetEnabled (false) ;
	return (RemoveHandler (info)) ? info : NULL ;
} ;

list_which LayerManager:: LayersWhich () const {
	return fLayersWhich ;
} ;

list_which LayerManager:: TargetsWhich () const {
	return fTargetsWhich ;
} ;